# src/pyneurorg/__init__.py

"""
pyneurorg: A Python module for simulating brain organoids using Brian2.
"""

# Define the version of the package
__version__ = "1.0.0"  # Incremente esta versão conforme você lança novas funcionalidades/correções

